@@ -11,7 +11,6 @@
 #include "pythonic/include/types/tuple.hpp"
 #include "pythonic/include/types/slice.hpp"
 #include "pythonic/include/types/vectorizable_type.hpp"
-#include "pythonic/include/types/nditerator.hpp"
 
 #include <ostream>
 #include <vector>
@@ -105,11 +104,11 @@ namespace types
         !std::is_same<contiguous_normalized_slice, S>::value;
 
     using shape_t = types::array<long, value>;
-    shape_t shape() const
+    template <size_t I>
+    auto shape() const
+        -> decltype(details::extract_shape(*this, utils::int_<I>{}))
     {
-      shape_t res;
-      details::init_shape(res, *this, utils::int_<value>{});
-      return res;
+      return details::extract_shape(*this, utils::int_<I>{});
     }
 
     // constructor
@@ -139,14 +138,24 @@ namespace types
     long size() const;
     explicit operator bool() const;
 
-    // accessor
-    T const &fast(long i) const;
-    T const &operator[](long i) const;
-    T &operator[](long i);
+    // accessors
+    const_reference fast(long i) const;
+    const_reference operator[](long i) const;
+    reference operator[](long i);
     sliced_list<T, S> operator[](contiguous_slice s) const;
     sliced_list<T, decltype(std::declval<S>() * std::declval<slice>())>
     operator[](slice s) const;
 
+    template <class... Indices>
+    dtype load(long index0, long index1, Indices... indices) const
+    {
+      return fast(index0).load(index1, indices...);
+    }
+
+    dtype load(long index) const
+    {
+      return fast(index);
+    }
     // comparison
     template <class K>
     bool operator==(list<K> const &other) const;
@@ -160,6 +169,16 @@ namespace types
     template <class vectorizer>
     simd_iterator vend(vectorizer) const;
 #endif
+
+    // other operations
+    template <class V>
+    bool contains(V const &v) const;
+    intptr_t id() const;
+
+    long count(T const &x) const;
+    template <class Tp, class Sp>
+    friend std::ostream &operator<<(std::ostream &os,
+                                    sliced_list<Tp, Sp> const &v);
   };
 
   /* list */
@@ -268,7 +287,10 @@ namespace types
     const_reverse_iterator rend() const;
 
     // comparison
-    int operator<(list<T> const &other) const;
+    bool operator<(list<T> const &other) const;
+    bool operator<=(list<T> const &other) const;
+    bool operator>(list<T> const &other) const;
+    bool operator>=(list<T> const &other) const;
 
 // element access
 #ifdef USE_XSIMD
@@ -289,6 +311,17 @@ namespace types
     sliced_list<T, contiguous_slice>
     operator[](contiguous_slice const &s) const;
 
+    template <class... Indices>
+    dtype load(long index0, long index1, Indices... indices) const
+    {
+      return fast(index0).load(index1, indices...);
+    }
+
+    dtype load(long index) const
+    {
+      return fast(index);
+    }
+
     // modifiers
     template <class Tp>
     void push_back(Tp &&x);
@@ -321,11 +354,10 @@ namespace types
 
     list<T> operator+(empty_list const &) const;
     list<T> operator*(long t) const;
+    list<T> const &operator*=(long t);
 
     template <class F>
-    list<T> &operator+=(list<F> const &s);
-    template <class Tp, size_t N, class V>
-    list<T> &operator+=(array_base<Tp, N, V> const &s);
+    list<T> &operator+=(F const &s);
 
     long size() const;
     template <class E>
@@ -340,14 +372,38 @@ namespace types
 
     long count(T const &x) const;
     using shape_t = array<long, value>;
-    shape_t shape() const
+    template <size_t I>
+    long shape() const
     {
-      shape_t res;
-      details::init_shape(res, *this, utils::int_<value>{});
+      if (I == 0)
+        return size();
+      else
+        return details::extract_shape(*this, utils::int_<I>{});
+    }
+
+    template <class Tp, size_t N, class V>
+    operator array_base<Tp, N, V>() const
+    {
+      assert(size() == N && "consistent size");
+      array_base<Tp, N, V> res;
+      std::copy(begin(), end(), res.begin());
       return res;
     }
   };
 
+  template <class T, size_t N>
+  list<T> operator*(static_list<T, N> const &self, long t)
+  {
+    list<T> res(self);
+    res *= t;
+    return res;
+  }
+  template <class T, size_t N>
+  list<T> operator*(long t, static_list<T, N> const &self)
+  {
+    return self * t;
+  }
+
   template <class T0, size_t N, class T1>
   list<typename __combined<T0, T1>::type>
   operator+(static_list<T0, N> const &l0, list<T1> const &l1)
@@ -379,14 +435,19 @@ namespace types
     template <class T, size_t N, class V>
     static_list<T, N> operator+(array_base<T, N, V> const &s) const;
     empty_list operator+(empty_list const &) const;
+    template <class F>
+    typename std::enable_if<!is_numexpr_arg<F>::value,
+                            list<typename F::value_type>>::type
+    operator+(F s) const;
     explicit operator bool() const;
     template <class T>
     operator list<T>() const;
     static constexpr long size();
 
-    shape_t shape() const
+    template <size_t I>
+    std::integral_constant<long, 0> shape() const
     {
-      return {0};
+      return {};
     }
 
     char fast(long) const
