@@ -7,7 +7,7 @@
 #include "pythonic/types/empty_iterator.hpp"
 #include "pythonic/types/attr.hpp"
 
-#include "pythonic/__builtin__/ValueError.hpp"
+#include "pythonic/builtins/ValueError.hpp"
 
 #include "pythonic/utils/nested_container.hpp"
 #include "pythonic/utils/shared_ref.hpp"
@@ -44,7 +44,7 @@
 #include "pythonic/utils/numpy_traits.hpp"
 #include "pythonic/utils/array_helper.hpp"
 
-#include "pythonic/__builtin__/len.hpp"
+#include "pythonic/builtins/len.hpp"
 #include "pythonic/operator_/iadd.hpp"
 #include "pythonic/operator_/iand.hpp"
 #include "pythonic/operator_/idiv.hpp"
@@ -121,13 +121,6 @@ namespace types
     return {std::move(self), i};
   }
 
-  template <class T, class pS>
-  long type_helper<ndarray<T, pS>>::step(ndarray<T, pS> const &self)
-  {
-    return std::accumulate(self.shape.begin() + 1, self.shape.end(), 1L,
-                           std::multiplies<long>());
-  }
-
   template <class T, class pS>
   typename type_helper<ndarray<T, pS> const &>::iterator
   type_helper<ndarray<T, pS> const &>::make_iterator(ndarray<T, pS> &n, long i)
@@ -165,13 +158,6 @@ namespace types
     return numpy_iexpr<ndarray<T, pS> const &>(self, i);
   }
 
-  template <class T, class pS>
-  long type_helper<ndarray<T, pS> const &>::step(ndarray<T, pS> const &self)
-  {
-    return std::accumulate(self.shape.begin() + 1, self.shape.end(), 1L,
-                           std::multiplies<long>());
-  }
-
   template <class T, class pS>
   typename type_helper<ndarray<T, pshape<pS>>>::iterator
   type_helper<ndarray<T, pshape<pS>>>::make_iterator(ndarray<T, pshape<pS>> &n,
@@ -206,13 +192,6 @@ namespace types
     return self.buffer[i];
   }
 
-  template <class T, class pS>
-  constexpr long
-  type_helper<ndarray<T, pshape<pS>>>::step(ndarray<T, pshape<pS>> const &)
-  {
-    return 1;
-  }
-
   template <class T, class pS>
   typename type_helper<ndarray<T, pshape<pS>> const &>::iterator
   type_helper<ndarray<T, pshape<pS>> const &>::make_iterator(
@@ -245,13 +224,6 @@ namespace types
     return self.buffer[i];
   }
 
-  template <class T, class pS>
-  constexpr long type_helper<ndarray<T, pshape<pS>> const &>::step(
-      ndarray<T, pshape<pS>> const &)
-  {
-    return 1;
-  }
-
   template <class T, class pS>
   typename type_helper<ndarray<T, array<pS, 1>>>::iterator
       type_helper<ndarray<T, array<pS, 1>>>::make_iterator(
@@ -285,13 +257,6 @@ namespace types
     return self.buffer[i];
   }
 
-  template <class T, class pS>
-  constexpr long type_helper<ndarray<T, array<pS, 1>>>::step(
-      ndarray<T, array<pS, 1>> const &)
-  {
-    return 1;
-  }
-
   template <class T, class pS>
   typename type_helper<ndarray<T, array<pS, 1>> const &>::iterator
       type_helper<ndarray<T, array<pS, 1>> const &>::make_iterator(
@@ -324,49 +289,68 @@ namespace types
     return self.buffer[i];
   }
 
-  template <class T, class pS>
-  constexpr long type_helper<ndarray<T, array<pS, 1>> const &>::step(
-      ndarray<T, array<pS, 1>> const &)
+  template <class S>
+  long patch_index(long index, S const &)
   {
-    return 1;
+    return index;
+  }
+  long patch_index(long index, std::integral_constant<long, 1> const &)
+  {
+    return 0;
   }
 
   template <size_t L>
-  template <class Ty, size_t M>
-  long noffset<L>::operator()(array<long, M> const &strides,
+  template <class S, class Ty, size_t M>
+  long noffset<L>::operator()(S const &strides,
                               array<Ty, M> const &indices) const
   {
-    return noffset<L - 1>{}(strides, indices) + strides[M - L] * indices[M - L];
+    auto index = patch_index(
+        indices[M - L],
+        typename std::tuple_element<M - L, typename S::shape_t>::type());
+    return noffset<L - 1>{}(strides, indices) +
+           strides.template strides<M - L>() * index;
   }
 
   template <size_t L>
-  template <class Ty, size_t M, class pS>
-  long noffset<L>::operator()(array<long, M> const &strides,
-                              array<Ty, M> const &indices,
+  template <class S, class Ty, size_t M, class pS>
+  long noffset<L>::operator()(S const &strides, array<Ty, M> const &indices,
                               pS const &shape) const
   {
+    auto index = patch_index(
+        indices[M - L],
+        typename std::tuple_element<M - L, typename S::shape_t>::type());
+    if (index < 0)
+      index += std::get<M - L>(shape);
+    assert(0 <= index and index < std::get<M - L>(shape));
     return noffset<L - 1>{}(strides, indices, shape) +
-           strides[M - L] * ((indices[M - L] < 0)
-                                 ? indices[M - L] + std::get<M - L>(shape)
-                                 : indices[M - L]);
+           strides.template strides<M - L>() *
+               ((index < 0) ? index + std::get<M - L>(shape) : index);
   }
 
   template <>
-  template <class Ty, size_t M>
-  long noffset<1>::operator()(array<long, M> const &,
+  template <class S, class Ty, size_t M>
+  long noffset<1>::operator()(S const &strides,
                               array<Ty, M> const &indices) const
   {
-    return indices[M - 1];
+    auto index = patch_index(
+        indices[M - 1],
+        typename std::tuple_element<M - 1, typename S::shape_t>::type());
+    return strides.template strides<M - 1>() * index;
   }
 
   template <>
-  template <class Ty, size_t M, class pS>
-  long noffset<1>::operator()(array<long, M> const &,
-                              array<Ty, M> const &indices,
+  template <class S, class Ty, size_t M, class pS>
+  long noffset<1>::operator()(S const &strides, array<Ty, M> const &indices,
                               pS const &shape) const
   {
-    return (indices[M - 1] < 0) ? indices[M - 1] + std::get<M - 1>(shape)
-                                : indices[M - 1];
+    auto index = patch_index(
+        indices[M - 1],
+        typename std::tuple_element<M - 1, typename S::shape_t>::type());
+    if (index < 0)
+      index += std::get<M - 1>(shape);
+    assert(0 <= index && index < std::get<M - 1>(shape));
+    return strides.template strides<M - 1>() *
+           ((index < 0) ? index + std::get<M - 1>(shape) : index);
   }
 
   /* constructors */
@@ -407,7 +391,7 @@ namespace types
   /* from a seed */
   template <class T, class pS>
   ndarray<T, pS>::ndarray(pS const &shape, none_type init)
-      : mem(sutils::prod(shape)), buffer(mem->data), _shape(shape),
+      : mem(sutils::sprod(shape)), buffer(mem->data), _shape(shape),
         _strides(make_strides(shape))
   {
   }
@@ -480,8 +464,8 @@ namespace types
   template <class T, class pS>
   template <class Op, class... Args>
   ndarray<T, pS>::ndarray(numpy_expr<Op, Args...> const &expr)
-      : mem(expr.flat_size()), buffer(mem->data), _shape(expr.shape()),
-        _strides(make_strides(_shape))
+      : mem(expr.flat_size()), buffer(mem->data),
+        _shape(sutils::getshape(expr)), _strides(make_strides(_shape))
   {
     initialize_from_expr(expr);
   }
@@ -489,8 +473,8 @@ namespace types
   template <class T, class pS>
   template <class Arg>
   ndarray<T, pS>::ndarray(numpy_texpr<Arg> const &expr)
-      : mem(expr.flat_size()), buffer(mem->data), _shape(expr.shape()),
-        _strides(make_strides(_shape))
+      : mem(expr.flat_size()), buffer(mem->data),
+        _shape(sutils::getshape(expr)), _strides(make_strides(_shape))
   {
     initialize_from_expr(expr);
   }
@@ -498,8 +482,8 @@ namespace types
   template <class T, class pS>
   template <class Arg>
   ndarray<T, pS>::ndarray(numpy_texpr_2<Arg> const &expr)
-      : mem(expr.flat_size()), buffer(mem->data), _shape(expr.shape()),
-        _strides(make_strides(_shape))
+      : mem(expr.flat_size()), buffer(mem->data),
+        _shape(sutils::getshape(expr)), _strides(make_strides(_shape))
   {
     initialize_from_expr(expr);
   }
@@ -507,8 +491,8 @@ namespace types
   template <class T, class pS>
   template <class Arg, class... S>
   ndarray<T, pS>::ndarray(numpy_gexpr<Arg, S...> const &expr)
-      : mem(expr.flat_size()), buffer(mem->data), _shape(expr.shape()),
-        _strides(make_strides(_shape))
+      : mem(expr.flat_size()), buffer(mem->data),
+        _shape(sutils::getshape(expr)), _strides(make_strides(_shape))
   {
     initialize_from_expr(expr);
   }
@@ -516,8 +500,8 @@ namespace types
   template <class T, class pS>
   template <class Arg>
   ndarray<T, pS>::ndarray(numpy_iexpr<Arg> const &expr)
-      : mem(expr.flat_size()), buffer(mem->data), _shape(expr.shape()),
-        _strides(make_strides(_shape))
+      : mem(expr.flat_size()), buffer(mem->data),
+        _shape(sutils::getshape(expr)), _strides(make_strides(_shape))
   {
     initialize_from_expr(expr);
   }
@@ -525,16 +509,8 @@ namespace types
   template <class T, class pS>
   template <class Arg, class F>
   ndarray<T, pS>::ndarray(numpy_vexpr<Arg, F> const &expr)
-      : mem(expr.flat_size()), buffer(mem->data), _shape(expr.shape()),
-        _strides(make_strides(_shape))
-  {
-    initialize_from_expr(expr);
-  }
-  template <class T, class pS>
-  template <class Arg>
-  ndarray<T, pS>::ndarray(fast_range<Arg> const &expr)
-      : mem(expr.flat_size()), buffer(mem->data), _shape(expr.shape()),
-        _strides(make_strides(_shape))
+      : mem(expr.flat_size()), buffer(mem->data),
+        _shape(sutils::getshape(expr)), _strides(make_strides(_shape))
   {
     initialize_from_expr(expr);
   }
@@ -618,7 +594,8 @@ namespace types
   typename std::enable_if<std::is_integral<Ty>::value, T &>::type
   ndarray<T, pS>::fast(array<Ty, value> const &indices)
   {
-    return *(buffer + noffset<std::tuple_size<pS>::value>{}(_strides, indices));
+    assert(inbound_indices(indices));
+    return *(buffer + noffset<std::tuple_size<pS>::value>{}(*this, indices));
   }
 
   template <class T, class pS>
@@ -626,7 +603,8 @@ namespace types
   typename std::enable_if<std::is_integral<Ty>::value, T>::type
   ndarray<T, pS>::fast(array<Ty, value> const &indices) const
   {
-    return *(buffer + noffset<std::tuple_size<pS>::value>{}(_strides, indices));
+    assert(inbound_indices(indices));
+    return *(buffer + noffset<std::tuple_size<pS>::value>{}(*this, indices));
   }
 
   template <class T, class pS>
@@ -656,7 +634,7 @@ namespace types
       operator[](array<Ty, value> const &indices) const
   {
     return *(buffer +
-             noffset<std::tuple_size<pS>::value>{}(_strides, indices, _shape));
+             noffset<std::tuple_size<pS>::value>{}(*this, indices, _shape));
   }
 
   template <class T, class pS>
@@ -666,7 +644,7 @@ namespace types
       operator[](array<Ty, value> const &indices)
   {
     return *(buffer +
-             noffset<std::tuple_size<pS>::value>{}(_strides, indices, _shape));
+             noffset<std::tuple_size<pS>::value>{}(*this, indices, _shape));
   }
 
   template <class T, class pS>
@@ -715,7 +693,7 @@ namespace types
   {
     sutils::push_front_t<pS, std::integral_constant<long, 1>> new_shape;
     sutils::copy_shape<1, -1>(
-        new_shape, _shape,
+        new_shape, *this,
         utils::make_index_sequence<std::tuple_size<pS>::value>());
     return reshape(new_shape);
   }
@@ -757,6 +735,15 @@ namespace types
     return extended_slice<count_new_axis<S0, S...>::value>{}((*this), s0, s...);
   }
 
+  template <class T, class pS>
+      template <class S0, class... S>
+      auto ndarray<T, pS>::operator()(S0 const &s0, S const &... s) &
+      -> decltype(extended_slice<count_new_axis<S0, S...>::value>{}((*this), s0,
+                                                                    s...))
+  {
+    return extended_slice<count_new_axis<S0, S...>::value>{}((*this), s0, s...);
+  }
+
   template <class T, class pS>
       template <class S0, class... S>
       auto ndarray<T, pS>::operator()(S0 const &s0, S const &... s) &&
@@ -777,7 +764,7 @@ namespace types
       numpy_vexpr<ndarray<T, pS>, ndarray<long, pshape<long>>>>::type
   ndarray<T, pS>::fast(F const &filter) const
   {
-    long sz = std::get<0>(filter.shape());
+    long sz = filter.template shape<0>();
     long *raw = (long *)malloc(sz * sizeof(long));
     long n = 0;
     for (long i = 0; i < sz; ++i)
@@ -851,7 +838,7 @@ namespace types
   }
 
   template <class T, class pS>
-  template <class Ty0, class Ty1, class... Tys>
+  template <class Ty0, class Ty1, class... Tys, class _>
   auto ndarray<T, pS>::
   operator[](std::tuple<Ty0, Ty1, Tys...> const &indices) const ->
       typename std::enable_if<
@@ -915,7 +902,7 @@ namespace types
   template <class T, class pS>
   long ndarray<T, pS>::flat_size() const
   {
-    return sutils::prod(_shape);
+    return sutils::prod(*this);
   }
   template <class T, class pS>
   bool ndarray<T, pS>::may_overlap(ndarray const &expr) const
@@ -940,7 +927,7 @@ namespace types
   template <class T, class pS>
   ndarray<T, pS>::operator bool() const
   {
-    if (sutils::any_of(_shape, [](long n) { return n != 1; }))
+    if (sutils::any_of(*this, [](long n) { return n != 1; }))
       throw ValueError("The truth value of an array with more than one element "
                        "is ambiguous. Use a.any() or a.all()");
     return *buffer;
@@ -955,7 +942,7 @@ namespace types
   template <class T, class pS>
   ndarray<T, pS> ndarray<T, pS>::copy() const
   {
-    ndarray<T, pS> res(_shape, __builtin__::None);
+    ndarray<T, pS> res(_shape, builtins::None);
     std::copy(fbegin(), fend(), res.fbegin());
     return res;
   }
@@ -966,12 +953,6 @@ namespace types
     return reinterpret_cast<intptr_t>(&(*mem));
   }
 
-  template <class T, class pS>
-  pS const &ndarray<T, pS>::shape() const
-  {
-    return _shape;
-  }
-
   /* pretty printing { */
   namespace impl
   {
@@ -1008,13 +989,12 @@ namespace types
   std::ostream &operator<<(std::ostream &os, ndarray<T, pS> const &e)
   {
     std::array<long, std::tuple_size<pS>::value> strides;
-    auto shape = e.shape();
+    auto shape = sutils::getshape(e);
     strides[std::tuple_size<pS>::value - 1] =
         std::get<std::tuple_size<pS>::value - 1>(shape);
     if (strides[std::tuple_size<pS>::value - 1] == 0)
       return os << "[]";
-    auto ashape = sutils::array(shape);
-    std::transform(strides.rbegin(), strides.rend() - 1, ashape.rbegin() + 1,
+    std::transform(strides.rbegin(), strides.rend() - 1, shape.rbegin() + 1,
                    strides.rbegin() + 1, std::multiplies<long>());
     size_t depth = std::tuple_size<pS>::value;
     int step = -1;
@@ -1095,7 +1075,7 @@ namespace std
 /* pythran attribute system { */
 #include "pythonic/numpy/transpose.hpp"
 PYTHONIC_NS_BEGIN
-namespace __builtin__
+namespace builtins
 {
   namespace details
   {
@@ -1130,7 +1110,7 @@ namespace __builtin__
             types::slice()))
     {
       using stype = typename types::is_complex<typename E::dtype>::type;
-      auto new_shape = sutils::array(a.shape());
+      auto new_shape = sutils::getshape(a);
       std::get<E::value - 1>(new_shape) *= 2;
       // this is tricky && dangerous!
       auto translated_mem =
@@ -1161,7 +1141,7 @@ namespace __builtin__
       // cannot use numpy.zero: forward declaration issue
       return {
           (typename E::dtype *)calloc(a.flat_size(), sizeof(typename E::dtype)),
-          a.shape(), types::ownership::owned};
+          sutils::getshape(a), types::ownership::owned};
     }
 
     template <class Op, class... Args>
@@ -1185,7 +1165,7 @@ namespace __builtin__
             types::slice()))
     {
       using stype = typename types::is_complex<typename E::dtype>::type;
-      auto new_shape = sutils::array(a.shape());
+      auto new_shape = sutils::getshape(a);
       std::get<E::value - 1>(new_shape) *= 2;
       // this is tricky && dangerous!
       auto translated_mem =
@@ -1199,10 +1179,9 @@ namespace __builtin__
   }
 
   template <class E>
-  auto getattr(types::attr::SHAPE, E const &a)
-      -> decltype(sutils::array(a.shape()))
+  types::array<long, E::value> getattr(types::attr::SHAPE, E const &a)
   {
-    return sutils::array(a.shape());
+    return sutils::getshape(a);
   }
 
   template <class E>
@@ -1216,7 +1195,7 @@ namespace __builtin__
   {
     types::array<long, E::value> strides;
     strides[E::value - 1] = sizeof(typename E::dtype);
-    auto shape = sutils::array(a.shape());
+    auto shape = sutils::getshape(a);
     std::transform(strides.rbegin(), strides.rend() - 1, shape.rbegin(),
                    strides.rbegin() + 1, std::multiplies<long>());
     return strides;
@@ -1381,12 +1360,12 @@ to_python<types::ndarray<T, pS>>::convert(types::ndarray<T, pS> const &cn,
           PyArray_DescrFromType(c_type_to_numpy_type<T>::value), nullptr);
     }
 
-    if (sutils::equals(n.shape(), pshape)) {
+    if (sutils::equals(n, pshape)) {
       if (transpose && !(PyArray_FLAGS(arr) & NPY_ARRAY_F_CONTIGUOUS))
         return PyArray_Transpose(arr, nullptr);
       else
         return p;
-    } else if (sutils::requals(n.shape(), pshape)) {
+    } else if (sutils::requals(n, pshape)) {
       if (transpose)
         return p;
       else
